import type { IDbUser } from "@interfaces/database/IDbUser";
import type { IReqLogin } from "@interfaces/requests/IReqLogin";
import type { IReqRegister } from "@interfaces/requests/IReqRegister";
import { ErrorType, type ISError } from "@interfaces/services/ISError";
import CredentialService from "@services/credential.service";
import JwtService from "@services/jwt.service";
import MySqlService from "@services/mysql.service";
import { Logger } from "tslog";
import { v4 } from "uuid";

const logger = new Logger({
	name: "UserService",
});

const DbHandler = new MySqlService.Handler("UserService");

/**
 * Retrieves a user from the database by the given email address.
 *
 * @param {string} targetEmail - The email address of the user to retrieve.
 * @returns {Promise<IDbUser | ISError>}
 * 			- A promise that resolves with the user.
 *   		- If the user is not found, an error object is returned.
 *   		- If an error occurs during the database operation, an error object is returned.
 */
async function getUserByEmail(targetEmail: string): Promise<IDbUser | ISError> {
	try {
		const dbUser = await MySqlService.User.getByEmail(DbHandler, targetEmail);
		if (dbUser === undefined) {
			logger.info(`\n\n> User not found (${targetEmail})\n`);
			return {
				error: ErrorType.NotFound,
				message: "The user was not fund.",
			};
		}
		return dbUser;
	} catch (err) {
		logger.error(err);
		return {
			error: ErrorType.DatabaseError,
			message: "An unknown error occurred.",
		};
	}
}

/**
 * Retrieves a user from the database based on the provided ID.
 *
 * @param {string} id - The ID of the user to retrieve.
 * @returns {Promise<IDbUser | ISError>} - A promise that resolves with the user object if found, or an error object if not.
 */
async function getUserFromIdService(id: string): Promise<IDbUser | ISError> {
	try {
		if (!id || id.length !== 36) {
			logger.info(`\n\n> Invalid ID (${id})\n`);
			return {
				error: ErrorType.InvalidData,
				message: "Invalid ID length.",
			};
		}
		const dbUser = await MySqlService.User.getById(DbHandler, id);
		if (dbUser === undefined) {
			logger.info(`User not found (${id})`);
			return {
				error: ErrorType.NotFound,
				message: "The user was not found.",
			};
		}
		return dbUser;
	} catch (err) {
		return {
			error: ErrorType.DatabaseError,
			message: "An unknown error occurred.",
		};
	}
}

/*
DbHandler.factorize({
	values: {
		id: '011010101',
		username: 'avnyr',
		age: 42,
		is_admin: true
	},
	actionName: "Testing"
}).then((result)=>{
	logger.trace(`\n\n> ${result._valuesArray.join(', ')}\n\n> ${result.totalFields}\n\n> ${result._keysTemplate}\n`)
})*/

//ToTest
/**
 * Registers a new user.
 *
 * @param {IReqRegister} inputData - The input data for registration.
 * @return {Promise<ISError | string>} - A Promise that resolves to either an error or a token.
 */
async function register(inputData: IReqRegister): Promise<ISError | string> {
	try {
		if (inputData.password.length < 6) {
			return {
				error: ErrorType.InvalidData,
				message: "Password must be at least 6 characters long.",
			};
		}

		const passwordHash = await CredentialService.hash(`${inputData.password}`);

		// Does the new user has accepted GDPR ?
		if (inputData.gdpr !== true) {
			return {
				error: ErrorType.InvalidData,
				message: "GDPR acceptance is required.",
			};
		}
		const currentDate = new Date();

		// Check if exist and return
		const dbUserIfExist: IDbUser | ISError = await getUserByEmail(
			inputData.email,
		);
		if ("error" in dbUserIfExist) {
			return {
				error: dbUserIfExist.error,
				message: dbUserIfExist.message,
			};
		}
		if (dbUserIfExist.id) {
			logger.info(
				`\n\n> User already exist for email "${inputData.email}".\n(${dbUserIfExist.username}::${dbUserIfExist.id})\n`,
			);
			return {
				error: ErrorType.UnAuthorized,
				message: "User already exists.",
			};
		}
		const currentId = v4();
		const NewUser = await MySqlService.User.insert(DbHandler, {
			id: currentId,
			email: inputData.email,
			username: inputData.username,
			firstname: inputData.firstName,
			lastname: inputData.lastName,
			dob: inputData.dob,
			hash: passwordHash,
			gdpr: currentDate,
			is_admin: false,
			is_mail_verified: false,
		});
		if ("error" in NewUser || NewUser.affectedRows === 0) {
			return {
				error: ErrorType.DatabaseError,
				message: "Error when inserting user in database.",
			};
		}
		logger.info(`\n\n> New user created ! (${inputData.username}::${currentId})\n`);

		// JWT
		const token = await JwtService.sign(
			{
				sub: currentId,
			},
			{
				alg: "HS512",
			},
			"1d",
			"user",
		);
		return token;
	} catch (err) {
		logger.error(`\n\n${err}\n`);
		return {
			error: ErrorType.DatabaseError,
			message: "An unknown error occurred.",
		};
	}
}

//ToTest
/**
 * Logs in a user with the provided input data.
 *
 * @param inputData - The input data for the login operation.
 *   @property email - The email of the user.
 *   @property password - The password of the user.
 * @returns A promise that resolves to either an error or a token string.
 *   @throws {ISError} - If an error occurs in the login process.
 *   @throws {string} - If the login was successful, returns a token string.
 */
async function login(inputData: IReqLogin): Promise<ISError | string> {
	try {
		const dbUser = await getUserByEmail(inputData.email);
		if ("error" in dbUser) {
			return {
				error: dbUser.error,
				message: dbUser.message,
			};
		}
		if (!dbUser.id) {
			return {
				error: ErrorType.NotFound,
				message: "User not found.",
			};
		}
		const isPasswordValid = await CredentialService.compare(
			inputData.password,
			dbUser.hash
		);
		if (!isPasswordValid) {
			return {
				error: ErrorType.UnAuthorized,
				message: "Invalid password.",
			};
		}
		const token = await JwtService.sign(
			{
				sub: dbUser.id,
				p: [{
					isAdmin: dbUser.is_admin,
					gdpr: dbUser.gdpr
				}]
			},
			{
				alg: "HS512",
			},
			"1d",
			"user"
		);
		return token;
	} catch (err) {
		logger.error(`\n\n${err}\n`);
		return {
			error: ErrorType.DatabaseError,
			message: "An unknown error occurred.",
		};
	}
}

//TOTest
/**
 * Retrieves all users from the database.
 *
 * @returns {Promise<Array<IDbUser> | ISError>} The list of users, or an error object if an error occurred.
 */
async function getAllUsersService(): Promise<Array<IDbUser> | ISError> {
	try {
		const allUsers = await MySqlService.User.getAll(DbHandler);
		if (allUsers === undefined) {
			logger.error(`Error retrieving all users.`);
			return {
				error: ErrorType.DatabaseError,
				message: "An unknown error occurred.",
			};
		}
		return allUsers;
	} catch (err) {
		logger.error(`\n\n${err}\n`);
    return {
      error: ErrorType.DatabaseError,
      message: "An unknown error occurred.",
    };
	}
}

async function editUserService(targetId, inputData: IDbUser): Promise<ISError | boolean> {
	if (!targetId || targetId.length !== 36) {
    logger.info(`\n\n> Invalid ID (${targetId})\n`);
    return {
      error: ErrorType.InvalidData,
      message: "Invalid ID length.",
    };
  }
	const dbUser = await MySqlService.User.getById(DbHandler, targetId)
	if (!dbUser.id) {
		return {
      error: ErrorType.NotFound,
      message: "User not found.",
    };
	}
	const result = await MySqlService.User.update(DbHandler, {
		username: inputData.username,
    firstname: inputData.firstname,
    lastname: inputData.lastname,
    dob: inputData.dob,
	})
}

/**
 * Delete a user from the database.
 *
 * @param {string} targetId - The ID of the user to be deleted.
 * @return {Promise<boolean>} - A promise that resolves to true if the user is successfully deleted, or false if an error occurs.
 */
async function deleteUserService(targetId) {
	logger.info(`Deleting user ${targetId}`);
	try {
		await Db.collection("users").deleteOne({
			id: targetId,
		});
		return true;
	} catch (e) {
		logger.warn(e);
		return false;
	}
}

logger.debug("\nService loaded.");

const UserService = {
	register: register,
	login: login,
	getAll: getAllUsersService,
	getFromId: getUserFromIdService,
	edit: editUserService,
	delete: deleteUserService,
};

export default UserService;
